/***********************************************************************/
/* This example code demonstrates a CAVE application program that      */
/*   uses the SGI Performer library to handle the visual rendering.    */
/*   For convenience, instead of the simple cube/pyramid world, this   */
/*   application loads a series of files in any format that Performer  */
/*   can handle.                                                       */
/*                                                                     */
/* The new features of this example include:                           */
/*   - The basic pfCAVE Library functions:                             */
/*      * pfInit()                                                     */
/*      * pfCAVEConfig()                                               */
/*      * pfInitChannels()                                             */
/*      * pfSync()                                                     */
/*      * pfCAVEPreFrame()                                             */
/*      * pfCAVEPostFrame()                                            */
/*      * pfCAVEHalt()                                                 */
/*      * pfExit()                                                     */
/*   - Some pfCAVE functions/variables used for interaction:           */
/*      * CAVEgetbutton() / CAVE_ESCKEY                                */
/*      * CAVEBUTTON1 ...                                              */
/*   - The pfCAVE functions used for travel:                           */
/*	* pfCAVEDCSNavTransform()                                      */
/*	* CAVENavLoadIdentity()                                        */
/*	* CAVENavRot()                                                 */
/*	* CAVEGetVector()                                              */
/*	* CAVENavTranslate()                                           */
/*   - Creates a single Scene Graph with all the object specified on   */
/*      the command line.                                              */
/*   - Exits the application when the 'Escape' key is pressed          */
/*   - Cleans up even when an interrupt signal is received             */
/*                                                                     */
/* Copyright 1998, University of Illinois Board of Trustees            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <pfcave.h>

/* This is very important -- these functions will be used as arguments. */
void	exit_application(int);
void	update_world();

static void     createScene(pfChannel * chan, int argc, char **argv);
static void     loadObjects(int argc, char **argv, pfGroup * parent);
static int      navigate(pfTraverser * trav, void *data);


int main(int argc, char **argv)
{
	/******************/
	/* setup the CAVE */
	signal(SIGINT, SIG_IGN);
	pfInit();
	signal(SIGINT, exit_application);
	pfCAVEConfig(&argc, argv, NULL);
	pfConfig();
	pfCAVEInitChannels();

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   a the expense of some performance.                        */
	if (CAVEConfig->Simulator)
		pfCAVEMasterChan()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);
	else	pfCAVEMasterChan()->getFStats()->setClass(PFSTATS_ALL, PFSTATS_OFF);

	/* */
	createScene(pfCAVEMasterChan(), argc, argv);
	while (!CAVEgetbutton(CAVE_ESCKEY)) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  Most Performer     */
		/* applications work using callback routines, which are    */
		/* stored in the scene-graph.                              */
		update_world();


		pfSync();
		/* latency critical operations go here */
		pfCAVEPreFrame();
		pfFrame();
		pfCAVEPostFrame();
	}

	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	pfCAVEHalt();
	pfExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine in most non-Performer */
/*   applications.  Most Performer apps can leave this empty.         */
void update_world()
{
	sginap(0);
}


/********************************************************************/
/* navigate(): calculates travels.  It is a callback function to    */
/*   alter the specified DCS.                                       */
/********************************************************************/
static int navigate(pfTraverser * trav, void *)
{
#define SPEED 2.0f
static	long	LastCAVETime;
	float	wand_point[3];
	float	joy_x, joy_y;
	float	delta_move;
	pfDCS	*navDcs = (pfDCS *) trav->getNode();	/* Get a pointer to the
							 * DCS node itself */

	joy_x = CAVE_JOYSTICK_X,
	joy_y = CAVE_JOYSTICK_Y;

	/* notice rotation for Performer is about the z-axis instead of y */
	if (fabs(joy_x) > 0.125)
		CAVENavRot((joy_x * -1.0), 'z');

	if (fabs(joy_y) > 0.125) {
		delta_move = joy_y * SPEED;
		CAVEGetVector(CAVE_WAND_FRONT, wand_point);
		CAVENavTranslate(wand_point[0] * delta_move,
			wand_point[1] * delta_move,
			wand_point[2] * delta_move);
	}

	/* Reset navigation on secret handshake */
	if ((CAVEBUTTON1) && (CAVEBUTTON3))
		CAVENavLoadIdentity();

	/* Load the DCS with the latest navigation matrix */
	pfCAVEDCSNavTransform(navDcs);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.). A DCS is added for */
/* navigation; the CAVE navigation transformation will be applied to the DCS,*/
/* so any scene nodes which are under the DCS will be drawn in world         */
/* coordinates.  A traversal function is used to update the DCS once per     */
/* frame, in the APP process. loadObjects() is then called to load any       */
/* objects given on the command line and place them under the DCS.           */
/*****************************************************************************/
static void createScene(pfChannel * chan, int argc, char **argv)
{
	pfScene        *scene;
	pfGeoState     *gstate;
	pfDCS          *dcs;

	scene = new pfScene;
	gstate = new pfGeoState;/* Set up a base geostate for the scene */
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/* Create a DCS for navigation */
	dcs = new pfDCS;

	scene->addChild(dcs);
	dcs->setTravFuncs(PFTRAV_APP, navigate, NULL);

	loadObjects(argc, argv, dcs);

	/* Assign the scene to the channels */
	chan->setScene(scene);
}


/******************************************************************************/
/* loadObjects(): loads objects specified on the command line (argc/argv).    */
/* pfdLoadFile() will load an object in any supported format; the file format */
/* is determined based on the extension (.iv, .obj, .pfb, etc.) If it         */
/* succeeds, it returns a pointer to the root of the subgraph containing the  */
/* object, which is then added to the scene under the given parent.  If it    */
/* fails, pfdLoadFile() returns NULL.                                         */
/******************************************************************************/
static void loadObjects(int argc, char **argv, pfGroup * parent)
{
	int             i;
	pfNode         *obj;
	for (i = 1; i < argc; i++) {
		/* Load the object */
		obj = pfdLoadFile(argv[i]);

		/* If it succeeded, add the object to the graph */
		if (obj)	
			parent->addChild(obj);
		else	pfExit();
	}
}

